package node

import (
	"bytes"
	"encoding/json"
	"log"
	"net"
	"strings"
	"sync"
	"time"

	"git.autistici.org/ale/autoradio"
	"git.autistici.org/ale/autoradio/instrumentation"
	"git.autistici.org/ale/autoradio/masterelection"
	"git.autistici.org/ale/autoradio/node/bwmonitor"
	"github.com/coreos/go-etcd/etcd"
)

var (
	icecastReloadErrors = instrumentation.NewCounter("icecast.reload_errors")
	icecastReloads      = instrumentation.NewCounter("icecast.reload")
	configIndex         = instrumentation.NewGauge("config.etcd_index")
)

func trigger(c chan bool) {
	select {
	case c <- true:
	default:
	}
}

// Converts an etcd key (a path) to the Icecast mount path. Removes
// mountPrefix from the beginning of the path, keeping the leading
// slash.
func keyToMount(key string) string {
	return key[len(autoradio.MountPrefix)-1:]
}

// In-memory representation of the overall configuration (basically
// just a list of the known mounts).
type clusterConfig struct {
	mounts map[string]*autoradio.Mount
	lock   sync.Mutex
}

func newClusterConfig() *clusterConfig {
	return &clusterConfig{
		mounts: make(map[string]*autoradio.Mount),
	}
}

// TODO: remove?
func (c *clusterConfig) GetMount(name string) *autoradio.Mount {
	c.lock.Lock()
	defer c.lock.Unlock()
	return c.mounts[name]
}

func (c *clusterConfig) ListMounts() []*autoradio.Mount {
	c.lock.Lock()
	defer c.lock.Unlock()
	result := make([]*autoradio.Mount, 0, len(c.mounts))
	for _, m := range c.mounts {
		result = append(result, m)
	}
	return result
}

// Update a mount (in-memory only).
func (c *clusterConfig) setMount(m *autoradio.Mount) {
	c.lock.Lock()
	defer c.lock.Unlock()
	c.mounts[m.Name] = m
}

// Delete a mount (in-memory only).
func (c *clusterConfig) delMount(name string) {
	c.lock.Lock()
	defer c.lock.Unlock()
	delete(c.mounts, name)
}

// Keeps the in-memory service configuration in sync with the etcd
// database. An update channel is triggered whenever the data changes.
type configWatcher struct {
	client autoradio.EtcdClient
	config *clusterConfig
	upch   chan bool
	stop   chan bool
	index  uint64
}

func newConfigSyncer(client autoradio.EtcdClient, config *clusterConfig, upch chan bool, stop chan bool) *configWatcher {
	return &configWatcher{
		client: client,
		config: config,
		upch:   upch,
		stop:   stop,
	}
}

func (w *configWatcher) setIndex(index uint64) {
	w.index = index
	configIndex.Set(int64(index))
}

func (w *configWatcher) updateConfigWithResponse(index uint64, key, value string) {
	mountName := keyToMount(key)
	log.Printf("updating mount %s [@%d]: %s", mountName, index, value)
	var m autoradio.Mount
	if err := json.NewDecoder(strings.NewReader(value)).Decode(&m); err != nil {
		log.Printf("corrupted data: %s: %s", value, err)
	} else {
		w.config.setMount(&m)
	}
}

func (w *configWatcher) watcher(ch chan *etcd.Response) {
	for {
		select {
		case response, ok := <-ch:
			if !ok {
				return
			}

			// Update the 'last seen' index, so that if
			// the Watcher dies, it knows where to start
			// from and we do not have to download the
			// full configuration again. Do this even for
			// actions we don't care about, just in case.
			w.setIndex(response.EtcdIndex)

			if response.Action == "delete" {
				mountName := keyToMount(response.Node.Key)
				log.Printf("deleted mount %s", mountName)
				w.config.delMount(mountName)
			} else if response.Action == "set" || response.Action == "create" || response.Action == "update" {
				w.updateConfigWithResponse(response.EtcdIndex, response.Node.Key, response.Node.Value)
			} else {
				continue
			}

			// Trigger an update.
			trigger(w.upch)

		case <-w.stop:
			return
		}
	}
}

// Load full configuration from etcd. This will trigger the update channel.
func (w *configWatcher) loadFullConfig() {
	for {
		response, err := w.client.Get(autoradio.MountPrefix, false, false)
		if err == nil && response.Node != nil && response.Node.Dir {
			// Directly update the configuration.
			for _, n := range response.Node.Nodes {
				w.updateConfigWithResponse(response.EtcdIndex, n.Key, n.Value)
			}
			log.Printf("got configuration at index %d", response.EtcdIndex)
			w.setIndex(response.EtcdIndex)
			break
		}
		log.Printf("Get error: %s", err)

		// Wait 1 second, but watch the stop channel.
		select {
		case <-time.After(1 * time.Second):
		case <-w.stop:
			return
		}
	}

	// Update the icecast daemon now that we have a full config.
	trigger(w.upch)
}

// Start the configWatcher in the background. It will wait for
// initialization to complete, so that when this function returns, the
// in-memory configuration has already been fully synchronized.
func (w *configWatcher) Start() {
	// Run until the first successful Get().
	log.Printf("retrieving initial config")
	w.loadFullConfig()

	// Main watch loop. Remember that etcd.Watch() will close the
	// receiver channel when it returns, so we need to start a new
	// syncer every time.
	go func() {
		for {
			ch := make(chan *etcd.Response)
			go w.watcher(ch)

			curIndex := w.index + 1
			log.Printf("starting watcher at index %d", curIndex)
			_, err := w.client.Watch(autoradio.MountPrefix, curIndex, true, ch, w.stop)
			if err == etcd.ErrWatchStoppedByUser {
				return
			} else if err != nil {
				// Log the error and start over.
				log.Printf("Watch(): %s", err)

				// If the error is code 401 ("index
				// out of date"), start from the
				// current index. TODO: consider
				// calling loadFullConfig, or we might
				// potentially lose some updates?
				if etcdErr, ok := err.(*etcd.EtcdError); ok && etcdErr.ErrorCode == 401 {
					w.setIndex(etcdErr.Index)
				}
			}
		}
	}()
}

// An active streaming node, managing the local icecast server.
type RadioNode struct {
	config *clusterConfig

	name      string
	ips       []net.IP
	client    autoradio.EtcdClient
	me        *masterelection.MasterElection
	watcher   *configWatcher
	icecast   Controller
	bw        *bwmonitor.BandwidthUsageMonitor
	heartbeat uint64
	upch      chan bool
	stop      chan bool
}

func NewRadioNode(name string, ips []net.IP, netDev string, bwLimit float64, client autoradio.EtcdClient) *RadioNode {
	config := newClusterConfig()

	// Network updates trigger icecast reconfiguration. This
	// channel is used as an 'event', no more than one entry will
	// be queued.
	upch := make(chan bool, 1)

	// MasterElection changes trigger an update.
	mech := make(chan masterelection.State)
	go func() {
		for _ = range mech {
			trigger(upch)
		}
	}()

	// Global 'stop' channel.
	stopch := make(chan bool)

	// Location information advertised when this node is master.
	minfo := &autoradio.MasterNodeInfo{
		Name: name,
		IP:   ips,
	}
	minfodata, err := json.Marshal(minfo)
	if err != nil {
		log.Fatal(err)
	}

	return &RadioNode{
		config: config,
		name:   name,
		ips:    ips,
		client: client,
		me: masterelection.NewMasterElection(
			client,
			autoradio.MasterElectionPath,
			string(minfodata),
			5,
			mech,
			stopch),
		watcher:   newConfigSyncer(client, config, upch, stopch),
		icecast:   NewIcecastController(name, stopch),
		heartbeat: 2,
		bw:        bwmonitor.NewBandwidthUsageMonitor(netDev, bwLimit),
		upch:      upch,
		stop:      stopch,
	}
}

// The presence goroutine periodically updates our entry in the list
// of nodes with the current node statistics.
func (rc *RadioNode) presence() {
	ticker := time.NewTicker(time.Duration(rc.heartbeat) * time.Second / 3)

	// Register ourselves using the node name.
	key := autoradio.NodePrefix + rc.name

	for {
		select {
		case <-ticker.C:
			// Build our NodeStatus.
			icecastStatus := rc.icecast.GetStatus()
			nodeStatus := autoradio.NodeStatus{
				Name:           rc.name,
				IP:             rc.ips,
				IcecastUp:      icecastStatus.Up,
				Mounts:         icecastStatus.Mounts,
				BandwidthUsage: rc.bw.GetUsage(),
			}

			// Update our node entry in the database.
			var buf bytes.Buffer
			json.NewEncoder(&buf).Encode(&nodeStatus)
			if _, err := rc.client.Set(key, buf.String(), rc.heartbeat); err != nil {
				log.Printf("presence: Set(): %s", err)
			}

		case <-rc.stop:
			return
		}
	}
}

// Get a valid IP address for the current master node (to be passed to
// Icecast). Since we don't really know much about network topology,
// just pick the first IP address associated with the master node.
func (rc *RadioNode) getMasterAddr() net.IP {
	var info autoradio.MasterNodeInfo
	if err := json.NewDecoder(strings.NewReader(rc.me.GetMasterData())).Decode(&info); err != nil || len(info.IP) == 0 {
		return nil
	}
	return info.IP[0]
}

// Run the node. This method does not return until someone calls Stop().
func (rc *RadioNode) Run() {
	// Bootstrap the config watcher. This ensures that we have a
	// full configuration (thanks to the Get() call) before we
	// start managing the icecast server.
	rc.watcher.Start()

	// Start the presence heartbeat.
	go rc.presence()

	// Start the masterelection runner.
	go rc.me.Run()

	// Wait an instant to give a chance to the other services to
	// initialize.
	time.Sleep(200 * time.Millisecond)

	// Start the Icecast status checker.
	go rc.icecast.Run()

	log.Printf("starting icecast updater")
	for {
		select {
		case <-rc.upch:
			icecastReloads.Incr()
			log.Printf("reloading icecast config")
			if err := rc.icecast.Update(rc.config, rc.me.IsMaster(), rc.getMasterAddr()); err != nil {
				icecastReloadErrors.Incr()
				log.Printf("Update(): %s", err)
			}

			// Do not reload icecast more often than once
			// every two seconds.
			time.Sleep(2 * time.Second)

		case <-rc.stop:
			return
		}
	}
}

// Stop everything.
func (rc *RadioNode) Stop() {
	close(rc.stop)

	// We should use WaitGroups here. Instead, wait 2 seconds.
	time.Sleep(2 * time.Second)
}
